#include "amxxmodule.h"
#include <curl\curl.h>
#include "CurlCleanerThread.h"
#include "CurlThread.h"

#ifdef _WIN32
#include <Windows.h>
#define SysSleep(ms) Sleep(ms);
#else
#include <unistd.h>
#define SysSleep(ms) usleep(ms*1000);
#endif

CCurlCleanerThread::CCurlCleanerThread() {
	m_bMustExit = false;
}

void CCurlCleanerThread::AddToRemoverQueue(CCurlThread* cct) {
	m_locker.Lock();

	if(m_bMustExit) {
		SERVER_PRINT("(CCurlCleanerThread::AddToRemoverQueue) CCurlCleanerThread object dead, can't add CCurlThread object to queue\n");
	}
	else {
		m_queueCurlThreads.push(cct);
	}
	m_locker.Unlock();
}

void CCurlCleanerThread::Exit() {
	m_locker.Lock();
	m_bMustExit = true;
	m_locker.Unlock();
}

void CCurlCleanerThread::Run() {
	while(true) {
		m_locker.Lock();

		while(!m_queueCurlThreads.empty()) {
			CCurlThread* cct = m_queueCurlThreads.front();

			cct->Wait();
			curl_easy_cleanup(cct->GetCurlHandle());
			delete cct;

			m_queueCurlThreads.pop();
		}

		if(m_bMustExit) {
			break;
		}

		m_locker.Unlock();

		SysSleep(300);
	}
}