#pragma once
#ifndef __CURLCALLBACKS_H__
#define __CURLCALLBACKS_H__

#include <map>
#include <curl\curl.h>
#include "amxxmodule.h"
#include "Thread.h"

using namespace std;

enum  { _FP_ARRAY_CELLS = 5, _FP_ARRAY_CELLSA, _FP_ARRAY_CHARS, _FP_ARRAY_CHARSA };

typedef map<CURLoption, int> TAMXCallbacks;
typedef map<CURLoption, cell> TAMXCBacksData;

struct TAmxCallbackParam {
	int type;
	int size;
	union {
		float	fData;
		cell	iData;
		void*	pData;
	};

	TAmxCallbackParam() {
		pData = NULL;
		size = 0;
		type = FP_CELL;
	}

	void Set(cell data) {
		this->iData = data;
	}

	void Set(void* data, int type, int size = 0) {
		this->pData = data;
		this->type = type;
		this->size = size;
	}

	void Setf(float data) {
		this->fData = data;
		this->type = FP_FLOAT;
	}
};

class CCurlThread : public CThread {
private:
	AMX*	m_pAMXPlugin;
	CURL*	m_pCURLHandle;
	
	TAMXCBacksData	m_amxCBacksData;
	TAMXCallbacks	m_amxCallbacks;
	
	cell	m_amxCompliteCallback;
	cell*	m_amxCompliteCbackData;
	int		m_amxCompliteCbackDataSize;

	CThreadLocker m_locker;
	CConditionVariable m_cvReady;

	//   
	struct TExecuteCallbackData {
		TExecuteCallbackData() { callback = -1; }

		TAmxCallbackParam* params;	//      
		int		count;
		int		callback;			// amx 
		struct TResult {
			TResult() { ready = false; }
			bool	ready;
			cell	data;
		} result;					//  
	} m_executeCBackData;			//       ,   

	bool m_bMustExit;

public:
			CCurlThread(AMX* plugin, CURL* handle);
			~CCurlThread();
	void	SetAmxCompliteCallback(char* function, cell* data, int data_len);
	void	SetAmxCallback(CURLoption opt, char* func_name);
	int		GetAmxCallback(CURLoption opt);

	void	SetAmxCallbackData(CURLoption opt, cell data);
	cell	GetAmxCallbackData(CURLoption opt);
	
	void	ExecuteNextAmxCallback(); //  amx    
	void	SafeExit();	//     ,  ,    

	bool	IsCurlRunning();
	CURL*	GetCurlHandle() { return m_pCURLHandle; }

	//     
	cell	ExecuteAmxCallback(CURLoption opt, TAmxCallbackParam* params, int count); //  ,    
	cell	ExecuteAmxCallback(int callback, TAmxCallbackParam* params, int count, CURLoption opt); 

protected:
	void	Run();
};

#endif // __CURLCALLBACKS_H__