#ifndef __THREADPOSIX_H__
#define __THREADPOSIX_H__

#include <pthread.h>

class CThread {
	pthread_t h;

public:
	CThread() {}
	virtual	~CThread() {}
	
	bool Start() {
		int err = pthread_create(&h, NULL, CThread::ThreadRun, (void*)this);
		if(err) {
			switch(err) {
			case EAGAIN:
				// ñëèøêîì ìíîãî ïîòîêîâ
				break;
			case EINVAL:
				// íåäîïóñòèìûé àðãóìåíò èëè íåâåðíûé ðàçìåð ñòåêà
				break;
			case EPERM:
				// íåäîñòàòî÷íî ðåñóðñîâ èëè ïðàâ
				break;
			}

			return false;
		}

		return true;
	}

	void Wait() {
		pthread_join(m_pThread, NULL);	
	}

protected:
	void Exit() {
		pthread_exit(NULL);
	}

	virtual void Run() = 0;

private:
	CThread(const CThread &th);
	static void ThreadRun(void* arg) {
		((CThread*)arg)->Run();	
	}
};

class CThreadLocker {
	pthread_mutex_t h;

public:
	CThreadLocker() {
		h = PTHREAD_MUTEX_INITIALIZER;	
	}
	
	void Lock() {
		pthread_mutex_lock(&h);
	}
	
	bool TryLock() {
		pthread_mutex_trylock(&h);
	}

	void Unlock() {
		pthread_mutex_unlock(&h);
	}

	pthread_mutex_t* handle() {
		return &h;	
	}

private:
	CThreadLocker(const CThreadLocker &thl) {}
};

class CConditionVariable {
	pthread_cond_t h;

public:
	CConditionVariable() {
		h = PTHREAD_COND_INITIALIZER;	
	}

	void Wait(CThreadLocker &locker) {
		pthread_cond_wait(&h, locker.handle());	
	}

	void WakeOne() {
		pthread_cond_signal(&h);	
	}

	void WakeAll() {
		pthread_cond_broadcast(&h);	
	}

private:
	CConditionVariable(const CConditionVariable &cv) {}
};

#endif // __THREADPOSIX_H__
