#include <Windows.h>
#include <process.h>
#include <errno.h>
#include "amxxmodule.h"
#include "Thread.h"

void CThread::Wait() {
	WaitForSingleObject(h, INFINITE);
}

bool CThread::Start() {
	h = (HANDLE)_beginthread(CThread::ThreadRun, 0, (void*)this);
	
	if((uintptr_t)h == -1L) {
		errno_t err;
		_get_errno(&err);

		switch(err) {
		case EAGAIN:
			//   
			break;
		case EINVAL:
			//      
			break;
		case EACCES:
			//    
			break;
		}

		return false;
	}

	return true;
}

void CThread::Exit() {
	_endthread();
}

void CThread::ThreadRun(void* arg) {
	((CThread*)arg)->Run();
}

// CThreadLocker

CThreadLocker::CThreadLocker() {
	InitializeCriticalSection(&h);
}

CThreadLocker::~CThreadLocker() {
	DeleteCriticalSection(&h);
}

void CThreadLocker::Lock() {
	EnterCriticalSection(&h);
}

bool CThreadLocker::TryLock() {
	TryEnterCriticalSection(&h);
}

void CThreadLocker::Unlock() {
	LeaveCriticalSection(&h);
}

CRITICAL_SECTION* CThreadLocker::handle() {
	return &h;
}

// CConditionVariable

CConditionVariable::CConditionVariable() {
	InitializeConditionVariable(&h);
}

void CConditionVariable::Wait(CThreadLocker &locker) {
	SleepConditionVariableCS(&h, locker.handle(), INFINITE);
}

void CConditionVariable::WakeOne() {
	WakeConditionVariable(&h);
}

void CConditionVariable::WakeAll() {
	WakeAllConditionVariable(&h);
}