#include <map>
#include "amxxmodule.h"
#include <curl\curl.h>
#include "CurlThread.h"
#include "CurlCleanerThread.h"
#include "curl_utils.h"

using namespace std;

map<CURL*, CCurlThread*>	g_curlThreads;		//  CCurlThread   
map<CURLoption, void*>		g_curlCallbacks;	//     
CCurlCleanerThread			g_cctCleaner;		//       CCurlThread    curl

int g_iLen;

// params[1]		CURL* handle	cell
// params[2]		char* url		str
// params[3]						str ref | escaped url
// params[4]		int len			cell |   
static cell AMX_NATIVE_CALL amx_curl_easy_escape(AMX* amx, cell* params) {
	char* str = MF_GetAmxString(amx, params[2], 0, &g_iLen);
	char* escaped_str = curl_easy_escape((CURL*)params[1], str, g_iLen);
	
	MF_SetAmxString(amx, params[3], escaped_str, params[4]);

	curl_free(escaped_str);

	return 0;
}

// ret				CURL*			cell
static cell AMX_NATIVE_CALL amx_curl_easy_init(AMX* amx, cell* params) {
	CURL* curl = curl_easy_init();

	g_curlThreads[curl] = new CCurlThread(amx, curl);

	return (cell)curl;
}

// params[1]		CURL* handle	cell
static cell AMX_NATIVE_CALL amx_curl_easy_cleanup(AMX* amx, cell* params) {
	CURL* curl = (CURL*)params[1];

	if(g_curlThreads.count(curl)) {
		g_cctCleaner.AddToRemoverQueue(g_curlThreads[curl]); //   CCurlThread     (+   )

		g_curlThreads[curl] = NULL; //     
	}
	else {
		MF_LogError(amx, AMX_ERR_NATIVE, "Attention! Linked object CCurlThread for curl handle \"%d\" was not found. Memory leak possible!", (int)curl);
	}
	
	return 0;
}

// params[1]		CURL *curl		cell
// params[2]		CURLINFO info	cell
// params[3]		...				cell,str,float ref
// params[4]		int				len | for string in third param
// ret				CURLcode		cell
static cell AMX_NATIVE_CALL amx_curl_easy_getinfo(AMX* amx, cell* params) {
	CURLcode code;
	CURLINFO info = (CURLINFO)params[2];
	CURL*	 curl = (CURL*)params[1];

	int curlinfo_mask = info & CURLINFO_TYPEMASK;
	
	if(curlinfo_mask == CURLINFO_STRING) {
		char* str;
		code = curl_easy_getinfo(curl, info, &str);

		if(code == CURLE_OK) {
			MF_SetAmxString(amx, params[3], str, params[4]);
		}
	}
	else if(curlinfo_mask == CURLINFO_LONG) {
		long num;
		code = curl_easy_getinfo(curl, info, &num);

		cell* ret = MF_GetAmxAddr(amx, params[3]);
		*ret = num;
	}
	else if(curlinfo_mask == CURLINFO_DOUBLE) {
		double num;
		code = curl_easy_getinfo(curl, info, &num);

		cell* ret = MF_GetAmxAddr(amx, params[3]);
		*ret = amx_ftoc(num);
	}
	else if(curlinfo_mask == CURLINFO_SLIST) {
		curl_slist* csl;
		code = curl_easy_getinfo(curl, info, &csl);

		cell* ret = MF_GetAmxAddr(amx, params[3]);
		*ret = (cell)csl;
	}
	else {
		return -1;
	}

	return (cell)code;
}

// params[1]		CURL *handle		cell
// params[2]		int bitmask			cell
// ret				CURLcode			cell
static cell AMX_NATIVE_CALL amx_curl_easy_pause(AMX* amx, cell* params) {
	return (cell)curl_easy_pause((CURL*)params[1], (int)params[2]);
}

// params[1]		CURL *handle		cell
// params[2]							str | complite callback ( public callback(code) {} )
// params[3]							arr | data
// params[4]							cell | data len
static cell AMX_NATIVE_CALL amx_curl_easy_perform(AMX* amx, cell* params) {
	CURL* curl = (CURL*)params[1];

	if(!g_curlThreads.count(curl)) {
		MF_LogError(amx, AMX_ERR_NATIVE, "Linked object CCurlThread for curl handle \"%d\" was not found. Reject...", (int)curl);
		return 0;
	}
	CCurlThread* cct = g_curlThreads[curl];

	/*if(cct->IsRunning()) {
		MF_LogError(amx, AMX_ERR_NATIVE, "Linked thread for curl handle \"%d\" already running. Reject...", (int)curl);
		return 0;
	}*/

	int callback_len;
	char* callback = MF_GetAmxString(amx, params[2], 0, &callback_len);

	cell* data = NULL;
	int data_len = (int)params[4];

	if(data_len > 0) {
		data = new cell[data_len];
		MF_CopyAmxMemory(data, MF_GetAmxAddr(amx, params[3]), data_len);
	}

	cct->SetAmxCompliteCallback(callback, data, data_len);
	cct->Start();

	return 0;
}

// params[1]		CURL* curl			cell
// params[2]		char* buffer		arr
// params[3]		size_t len			cell
// params[4]		size_t* n			cell
// ret				CURLcode			cell
static cell AMX_NATIVE_CALL amx_curl_easy_recv(AMX* amx, cell* params) {
	size_t n;
	char* buf;
	CURLcode res = curl_easy_recv((CURL*)params[1], buf, (size_t)params[3], &n);

	cell* retn = MF_GetAmxAddr(amx, params[4]);
	*retn = (cell)n;

	MF_SetAmxString(amx, params[2], buf, params[3]);

	return res;
}

// params[1]		CURL* curl			cell
static cell AMX_NATIVE_CALL amx_curl_easy_reset(AMX* amx, cell* params) {
	CURL* curl = (CURL*)params[1];

	curl_easy_reset(curl);

	if(!g_curlThreads.count(curl)) {
		MF_LogError(amx, AMX_ERR_NATIVE, "Linked object CCurlThread for curl handle \"%d\" not found.", (int)curl);
		return 0;
	}

	delete g_curlThreads[curl];
	g_curlThreads[curl] = new CCurlThread(amx, curl);

	return 0;
}

// params[1]		CURL* curl			cell
// params[2]		char* buffer		arr
// params[3]		size_t len			cell
// params[4]		size_t* n			cell
// ret				CURLcode			cell
static cell AMX_NATIVE_CALL amx_curl_easy_send(AMX* amx, cell* params) {
	size_t n;
	CURLcode res = curl_easy_send((CURL*)params[1], MF_GetAmxString(amx, params[2], 0, &g_iLen), (size_t)params[3], &n);

	cell* retn = MF_GetAmxAddr(amx, params[4]);
	*retn = (cell)n;

	return res;
}

// params[1]		CURL *handle		cell
// params[2]		CURLoption option	cell
// params[3]		...					str,cell
// ret				CURLcode			cell
static cell AMX_NATIVE_CALL amx_curl_easy_setopt(AMX* amx, cell* params) {
	CURL* curl = (CURL*)params[1];
	CURLoption option = (CURLoption)params[2];

	if(!g_curlThreads.count(curl)) {
		MF_LogError(amx, AMX_ERR_NATIVE, "Linked object CCurlThread for curl handle \"%d\" was not found. Reject...", (int)curl);
		return -1;
	}
	CCurlThread* cct = g_curlThreads[curl];

	/*if(cct->IsRunning()) {
		MF_LogError(amx, AMX_ERR_NATIVE, "Linked thread for curl handle \"%d\" is running. Can't set/change option. Reject...", (int)curl);
		return -1;
	}*/

	if(option < CURLOPTTYPE_OBJECTPOINT) { // CURLOPTTYPE_LONG
		return curl_easy_setopt(curl, option, (long)*MF_GetAmxAddr(amx, params[3]));
	}
	else if(option < CURLOPT_WRITEFUNCTION) { // CURLOPTTYPE_OBJECTPOINT
		if(CurlIsDataOption(option)) {
			cct->SetAmxCallbackData(option, *MF_GetAmxAddr(amx, params[3]));
			return CURLE_OK;
		}
		else {
			switch(option) {
			case CURLOPT_HTTPPOST: 
				return curl_easy_setopt(curl, option, (curl_httppost*)*MF_GetAmxAddr(amx, params[3]));
			case CURLOPT_HTTPHEADER:
			case CURLOPT_PROXYHEADER:
			case CURLOPT_HTTP200ALIASES:
			case CURLOPT_MAIL_RCPT:
			case CURLOPT_QUOTE:
			case CURLOPT_POSTQUOTE:
			case CURLOPT_RESOLVE:
			case CURLOPT_TELNETOPTIONS:
				return curl_easy_setopt(curl, option, (curl_slist*)*MF_GetAmxAddr(amx, params[3]));
			}

			return curl_easy_setopt(curl, option, MF_GetAmxString(amx, params[3], 0, &g_iLen));
		}
	}
	else if(option < CURLOPTTYPE_OFF_T) { // CURLOPT_WRITEFUNCTION
		CURLoption dataOpt = CurlLinkedDataOption(option);
		if(dataOpt == CURLOPT_LASTENTRY) {
			MF_LogError(amx, AMX_ERR_NATIVE, "Option %d not supported. Reject...", (int)option);
			return -1;
		}

		char* func_name = MF_GetAmxString(amx, params[3], 0, &g_iLen);

		cct->SetAmxCallback(option, func_name);

		/*code = */curl_easy_setopt(curl, dataOpt, (void*)cct);
		return curl_easy_setopt(curl, option, g_curlCallbacks[option]);
	}
	else { // CURLOPTTYPE_OFF_T
		/*curl_off_t val = 0;
		cell* p = MF_GetAmxAddr(amx, params[3]);

		val = p[0];
		val <<= 32;
		val |= p[1];

		return curl_easy_setopt(curl, option, val);*/
		return curl_easy_setopt(curl, option, (curl_off_t)*MF_GetAmxAddr(amx, params[3]));
	}

	return -1; //   
}

// params[1]		CURLcode err		cell
// params[2]		char* strerr		str
// params[3]		int len				cell
// ret
static cell AMX_NATIVE_CALL amx_curl_easy_strerror(AMX* amx, cell* params) {
	MF_SetAmxString(amx, params[2], curl_easy_strerror((CURLcode)params[1]), params[3]);

	return 0;
}

// params[1]		CURL* handle		cell
// params[2]		char* url			str
// params[3]							str ref | escaped url
// params[4]		int len				cell |   
// ret				int new_len			cell |    
static cell AMX_NATIVE_CALL amx_curl_easy_unescape(AMX* amx, cell* params) {
	char* str = MF_GetAmxString(amx, params[2], 0, &g_iLen);

	int new_len;
	char* unescaped_str = curl_easy_unescape((CURL*)params[1], str, g_iLen, &new_len);
	MF_SetAmxString(amx, params[3], unescaped_str, params[4]);

	curl_free(unescaped_str);

	return new_len;
}

// params[1]		curl_httppost*first	cell
// params[2]		curl_httppost*last	cell
// ...
// ret				CURLFORMcode		cell
static cell AMX_NATIVE_CALL amx_curl_formadd(AMX* amx, cell* params) {
	curl_httppost* first = (curl_httppost*)*MF_GetAmxAddr(amx, params[1]);
	curl_httppost* last = (curl_httppost*)*MF_GetAmxAddr(amx, params[2]);

	int i = 3, pairs = 0;
	while((CURLformoption)*MF_GetAmxAddr(amx, params[i]) != CURLFORM_END) { i+=2; pairs++; }

	if(pairs == 0 || pairs > 14) {
		MF_LogError(amx, AMX_ERR_NATIVE, "Invalid params count, get %d, must be in range 0 - 14. Reject...", pairs);

		return -1;
	}

	char strings[14][16384];

	curl_forms* forms = new curl_forms[pairs + 1];
	for(i = 0; i < pairs; i++) {
		strcpy(strings[i], MF_GetAmxString(amx, params[i*2 + 4], 0, &g_iLen));

		forms[i].option = (CURLformoption)*MF_GetAmxAddr(amx, params[i*2 + 3]);
		forms[i].value = strings[i];
	}
	forms[pairs].option = CURLFORM_END;
	
	CURLFORMcode code = curl_formadd(&first, &last, CURLFORM_ARRAY, forms, CURLFORM_END);

	delete[pairs + 1] forms;

	return code;
}

// params[1]		curl_httppost*first	cell
static cell AMX_NATIVE_CALL amx_curl_formfree(AMX* amx, cell* params) {
	curl_formfree((curl_httppost*)params[1]);

	return 0;
}

// params[1]		curl_slist* list	cell
// params[2]		char* string		arr
// ret				curl_slist* list	cell
static cell AMX_NATIVE_CALL amx_curl_slist_append(AMX* amx, cell* params) {
	return (cell)curl_slist_append((curl_slist*)params[1], MF_GetAmxString(amx, params[2], 0, &g_iLen));
}

// params[1]		curl_slist* list	cell
static cell AMX_NATIVE_CALL amx_curl_slist_free_all(AMX* amx, cell* params) {
	curl_slist_free_all((curl_slist*)params[1]);
	
	return 0;
}

// params[1]		char* buf			str
// params[2]		int len				cell
// ret
static cell AMX_NATIVE_CALL amx_curl_version(AMX* amx, cell* params) {
	char* version = curl_version();

	MF_SetAmxString(amx, params[1], version, params[2]);

	return 0;
}

AMX_NATIVE_INFO g_AmxCurlNatives[] =
{
	{"curl_easy_escape",		amx_curl_easy_escape},
	{"curl_easy_init",			amx_curl_easy_init},
	{"curl_easy_cleanup",		amx_curl_easy_cleanup},
	{"curl_easy_getinfo",		amx_curl_easy_getinfo},
	{"curl_easy_pause",			amx_curl_easy_pause},
	{"curl_easy_perform",		amx_curl_easy_perform},
	{"curl_easy_recv",			amx_curl_easy_recv},
	{"curl_easy_reset",			amx_curl_easy_reset},
	{"curl_easy_send",			amx_curl_easy_send},
	{"curl_easy_setopt",		amx_curl_easy_setopt},
	{"curl_easy_strerror",		amx_curl_easy_strerror},
	{"curl_easy_unescape",		amx_curl_easy_unescape},
	{"curl_formadd",			amx_curl_formadd},
	{"curl_formfree",			amx_curl_formfree},
	{"curl_slist_append",		amx_curl_slist_append},
	{"curl_slist_free_all",		amx_curl_slist_free_all},
	{"curl_version",			amx_curl_version},
	{NULL,						NULL},
};