#include <openssl\rsa.h>
#include <openssl\engine.h>
#include "amxxmodule.h"

using namespace std;

// api :)

// params[1]						arr | pub modulus
// params[2]						arr | pub exponent
// params[3]						arr |   
// params[4]						arr |  
// params[5]						cell| 
// ret
static cell AMX_NATIVE_CALL amx_RSA_public_encrypt(AMX* amx, cell* params) {
	RSA* rsa = RSA_new();
	
	int mod_len;
	char* mod = MF_GetAmxString(amx, params[1], 0, &mod_len);

	int exp_len;
	char* exp = MF_GetAmxString(amx, params[2], 1, &exp_len);

	int enc_len;
	char* encrypt = MF_GetAmxString(amx, params[3], 2, &enc_len);

	rsa->n = BN_bin2bn((const unsigned char*)mod, mod_len, NULL);
	rsa->e = BN_bin2bn((const unsigned char*)exp, exp_len, NULL);

	int rsa_size = RSA_size(rsa);
	if(rsa_size > params[5]) {
		MF_LogError(amx, AMX_ERR_NATIVE, "encrypted array size to small (need %d cells)", rsa_size);
		
		RSA_free(rsa);
		return -1;
	}

	unsigned char* encrypted = new unsigned char[rsa_size];
	int size = RSA_public_encrypt(enc_len, (const unsigned char*)encrypt, encrypted, rsa, RSA_PKCS1_OAEP_PADDING);
	if(size < 0) {
		unsigned long err = ERR_get_error();
		MF_LogError(amx, AMX_ERR_NATIVE, "Error: %s\n%s\n", ERR_reason_error_string(err), ERR_error_string(err, NULL));
		
		delete[rsa_size] encrypted;
		RSA_free(rsa);
		return -1;
	}

	MF_SetAmxString(amx, params[4], (char*)encrypted, params[5]);
	
	delete[rsa_size] encrypted;
	RSA_free(rsa);
	return size;
}

AMX_NATIVE_INFO g_AmxOpenSSLNatives[] =
{
	{"_RSA_public_encrypt",		amx_RSA_public_encrypt},
	{NULL,						NULL}
};