#include <map>
#include <list>
#include "amxxmodule.h"
#include <curl\curl.h>
#include "CurlThread.h"
#include "curl_callbacks.h"
#include "amx_utils.h"

using namespace std;

// amx_natives.cpp
extern map<CURLoption, void*> g_curlCallbacks;

size_t write_callback(char *ptr, size_t size, size_t nmemb, void *userdata) {
	CCurlThread* cct = (CCurlThread*)userdata;

	TAmxCallbackParam params[4];
	params[0].Set(ptr, _FP_ARRAY_CHARS, (nmemb * size));
	params[1].Set(size);
	params[2].Set(nmemb);
	params[3].Set(cct->GetAmxCallbackData(CURLOPT_WRITEDATA));

	return cct->ExecuteAmxCallback(CURLOPT_WRITEFUNCTION, params, 4);
}

size_t read_callback(char *buffer, size_t size, size_t nitems, void *instream) {
	CCurlThread* cct = (CCurlThread*)instream;

	if(nitems > 1500) {
		nitems = 1500;
	}

	TAmxCallbackParam params[4];
	params[0].Set(buffer, _FP_ARRAY_CHARSA, (nitems * size));
	params[1].Set(size);
	params[2].Set(nitems);
	params[3].Set(cct->GetAmxCallbackData(CURLOPT_READDATA));

	return cct->ExecuteAmxCallback(CURLOPT_READFUNCTION, params, 4);
}

curlioerr ioctl_callback(CURL *handle, int cmd, void *clientp) {
	CCurlThread* cct = (CCurlThread*)clientp;

	TAmxCallbackParam params[3];
	params[0].Set((cell)handle);
	params[1].Set(cmd);
	params[2].Set(cct->GetAmxCallbackData(CURLOPT_IOCTLDATA));

	return (curlioerr)cct->ExecuteAmxCallback(CURLOPT_IOCTLFUNCTION, params, 3);
}

int seek_callback(void *userp, curl_off_t offset, int origin) {
	CCurlThread* cct = (CCurlThread*)userp;

	TAmxCallbackParam params[3];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_SEEKDATA));
	params[1].Set(offset);
	params[2].Set(origin);

	return cct->ExecuteAmxCallback(CURLOPT_SEEKFUNCTION, params, 3);
}

int sockopt_callback(void *clientp, curl_socket_t curlfd, curlsocktype purpose) {
	CCurlThread* cct = (CCurlThread*)clientp;

	TAmxCallbackParam params[3];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_SOCKOPTDATA));
	params[1].Set(curlfd);
	params[2].Set(purpose);

	return cct->ExecuteAmxCallback(CURLOPT_SOCKOPTFUNCTION, params, 3);
}

curl_socket_t opensocket_callback(void *clientp, curlsocktype purpose, struct curl_sockaddr *address) {
	CCurlThread* cct = (CCurlThread*)clientp;

	TAmxCallbackParam params[3];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_OPENSOCKETDATA));
	params[1].Set(purpose);
	params[2].Set((cell)address);

	return (curl_socket_t)cct->ExecuteAmxCallback(CURLOPT_OPENSOCKETFUNCTION, params, 3);
}

int closesocket_callback(void *clientp, curl_socket_t item) {
	CCurlThread* cct = (CCurlThread*)clientp;

	TAmxCallbackParam params[2];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_CLOSESOCKETDATA));
	params[1].Set(item);

	return cct->ExecuteAmxCallback(CURLOPT_CLOSESOCKETFUNCTION, params, 2);
}

int progress_callback(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow) {
	CCurlThread* cct = (CCurlThread*)clientp;

	TAmxCallbackParam params[5];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_PROGRESSDATA));
	params[1].Setf(dltotal);
	params[2].Setf(dlnow);
	params[3].Setf(ultotal);
	params[4].Setf(ulnow);

	return cct->ExecuteAmxCallback(CURLOPT_PROGRESSFUNCTION, params, 5);
}

int progress_callback_x(void *clientp, curl_off_t dltotal, curl_off_t dlnow, curl_off_t ultotal, curl_off_t ulnow) {
	CCurlThread* cct = (CCurlThread*)clientp;

	TAmxCallbackParam params[5];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_XFERINFODATA));
	params[1].Set(dltotal);
	params[2].Set(dlnow);
	params[3].Set(ultotal);
	params[4].Set(ulnow);

	return cct->ExecuteAmxCallback(CURLOPT_XFERINFOFUNCTION, params, 5);
}

size_t header_callback(char *buffer, size_t size, size_t nitems, void *userdata) {
	CCurlThread* cct = (CCurlThread*)userdata;

	TAmxCallbackParam params[4];
	params[0].Set(buffer, _FP_ARRAY_CHARS, (size * nitems));
	params[1].Set(size);
	params[2].Set(nitems);
	params[3].Set(cct->GetAmxCallbackData(CURLOPT_HEADERDATA));

	return (size_t)cct->ExecuteAmxCallback(CURLOPT_HEADERFUNCTION, params, 4);
}

int debug_callback(CURL *handle, curl_infotype type, char *data, size_t size, void *userptr) {
	CCurlThread* cct = (CCurlThread*)userptr;

	TAmxCallbackParam params[5];
	params[0].Set((cell)handle);
	params[1].Set(type);
	params[2].Set(data, _FP_ARRAY_CHARS, size);
	params[3].Set(size);
	params[4].Set(cct->GetAmxCallbackData(CURLOPT_DEBUGDATA));

	return cct->ExecuteAmxCallback(CURLOPT_DEBUGFUNCTION, params, 5);
}

CURLcode ssl_ctx_callback(CURL *curl, void *ssl_ctx, void *userptr) {
	CCurlThread* cct = (CCurlThread*)userptr;

	TAmxCallbackParam params[3];
	params[0].Set((cell)curl);
	params[1].Set((cell)ssl_ctx);
	params[2].Set(cct->GetAmxCallbackData(CURLOPT_SSL_CTX_DATA));

	return (CURLcode)cct->ExecuteAmxCallback(CURLOPT_SSL_CTX_FUNCTION, params, 3);
}

size_t interleave_callback(void *ptr, size_t size, size_t nmemb, void *userdata) { // @TODO
	CCurlThread* cct = (CCurlThread*)userdata;

	TAmxCallbackParam params[4];
	params[0].Set(ptr, _FP_ARRAY_CELLS, (size * nmemb));
	params[1].Set(size);
	params[2].Set(nmemb);
	params[3].Set(cct->GetAmxCallbackData(CURLOPT_INTERLEAVEDATA));

	return (size_t)cct->ExecuteAmxCallback(CURLOPT_INTERLEAVEFUNCTION, params, 4);
}

long chunk_bgn_callback(const void *transfer_info, void *ptr, int remains) {
	CCurlThread* cct = (CCurlThread*)ptr;

	TAmxCallbackParam params[3];
	params[0].Set((cell)transfer_info);
	params[1].Set(cct->GetAmxCallback(CURLOPT_CHUNK_BGN_FUNCTION));
	params[2].Set(remains);

	return (long)cct->ExecuteAmxCallback(CURLOPT_CHUNK_BGN_FUNCTION, params, 3);
}

long chunk_end_callback(void *ptr) {
	CCurlThread* cct = (CCurlThread*)ptr;

	TAmxCallbackParam params[1];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_CHUNK_DATA));

	return (long)cct->ExecuteAmxCallback(CURLOPT_CHUNK_END_FUNCTION, params, 1);
}

int fnmatch_callback(void *ptr, const char *pattern, const char *string) {
	CCurlThread* cct = (CCurlThread*)ptr;

	TAmxCallbackParam params[3];
	params[0].Set(cct->GetAmxCallbackData(CURLOPT_FNMATCH_DATA));
	params[1].Set((void*)pattern, FP_STRING);
	params[2].Set((void*)string, FP_STRING);
	
	return cct->ExecuteAmxCallback(CURLOPT_FNMATCH_FUNCTION, params, 3);
}

void CCB_InitializeCallbacksMap() {
	g_curlCallbacks[CURLOPT_WRITEFUNCTION]		= (void*)write_callback;
	g_curlCallbacks[CURLOPT_READFUNCTION]		= (void*)read_callback;
	g_curlCallbacks[CURLOPT_IOCTLFUNCTION]		= (void*)ioctl_callback;
	g_curlCallbacks[CURLOPT_SEEKFUNCTION]		= (void*)seek_callback;
	g_curlCallbacks[CURLOPT_SOCKOPTFUNCTION]	= (void*)sockopt_callback;
	g_curlCallbacks[CURLOPT_OPENSOCKETFUNCTION] = (void*)opensocket_callback;
	g_curlCallbacks[CURLOPT_CLOSESOCKETFUNCTION]= (void*)closesocket_callback;
	g_curlCallbacks[CURLOPT_PROGRESSFUNCTION]	= (void*)progress_callback;
	g_curlCallbacks[CURLOPT_XFERINFOFUNCTION]	= (void*)progress_callback_x;
	g_curlCallbacks[CURLOPT_HEADERFUNCTION]		= (void*)header_callback;
	g_curlCallbacks[CURLOPT_DEBUGFUNCTION]		= (void*)debug_callback;
	g_curlCallbacks[CURLOPT_SSL_CTX_FUNCTION]	= (void*)ssl_ctx_callback;
	g_curlCallbacks[CURLOPT_INTERLEAVEFUNCTION] = (void*)interleave_callback;
	g_curlCallbacks[CURLOPT_CHUNK_BGN_FUNCTION] = (void*)chunk_bgn_callback;
	g_curlCallbacks[CURLOPT_CHUNK_END_FUNCTION] = (void*)chunk_end_callback;
	g_curlCallbacks[CURLOPT_FNMATCH_FUNCTION]	= (void*)fnmatch_callback;
}

/*struct TCurlCallback {
	CURLoption	callbackOption;
	CURLoption	dataOption;
	void*		function;
	int			instantError;

	TCurlCallback(CURLoption opt, CURLoption data_opt, void* func, int inst_err) {
		callbackOption = opt;
		dataOption = data_opt;
		function = func;
		instantError = inst_err;
	}
};*/